@@ -5,8 +5,10 @@ from datetime import datetime
 import hashlib
 import pathlib
 import sqlite_utils
-from sqlite_utils.db import AlterError, DescIndex
+from sqlite_utils.db import AlterError, BadMultiValues, DescIndex
+from sqlite_utils import recipes
 import textwrap
+import inspect
 import io
 import itertools
 import json
@@ -15,10 +17,12 @@ import sys
 import csv as csv_std
 import tabulate
 from .utils import (
+    _compile_code,
     file_progress,
     find_spatialite,
     sqlite3,
     decode_base64_values,
+    progressbar,
     rows_from_file,
     Format,
     TypeTracker,
@@ -41,7 +45,7 @@ It's often worth trying: --encoding=latin-1
 """.strip()
 
 
-# Increase CSV field size limit to maximim possible
+# Increase CSV field size limit to maximum possible
 # https://stackoverflow.com/a/15063941
 field_size_limit = sys.maxsize
 
@@ -635,13 +639,36 @@ def insert_upsert_options(fn):
                 required=True,
             ),
             click.argument("table"),
-            click.argument("json_file", type=click.File("rb"), required=True),
+            click.argument("file", type=click.File("rb"), required=True),
             click.option(
                 "--pk", help="Columns to use as the primary key, e.g. id", multiple=True
             ),
+            click.option(
+                "--flatten",
+                is_flag=True,
+                help='Flatten nested JSON objects, so {"a": {"b": 1}} becomes {"a_b": 1}',
+            ),
             click.option("--nl", is_flag=True, help="Expect newline-delimited JSON"),
-            click.option("-c", "--csv", is_flag=True, help="Expect CSV"),
-            click.option("--tsv", is_flag=True, help="Expect TSV"),
+            click.option("-c", "--csv", is_flag=True, help="Expect CSV input"),
+            click.option("--tsv", is_flag=True, help="Expect TSV input"),
+            click.option(
+                "--lines",
+                is_flag=True,
+                help="Treat each line as a single value called 'line'",
+            ),
+            click.option(
+                "--text",
+                is_flag=True,
+                help="Treat input as a single value called 'text'",
+            ),
+            click.option("--convert", help="Python code to convert each item"),
+            click.option(
+                "--import",
+                "imports",
+                type=str,
+                multiple=True,
+                help="Python modules to import",
+            ),
             click.option("--delimiter", help="Delimiter to use for CSV files"),
             click.option("--quotechar", help="Quote character to use for CSV/TSV"),
             click.option(
@@ -691,11 +718,16 @@ def insert_upsert_options(fn):
 def insert_upsert_implementation(
     path,
     table,
-    json_file,
+    file,
     pk,
+    flatten,
     nl,
     csv,
     tsv,
+    lines,
+    text,
+    convert,
+    imports,
     delimiter,
     quotechar,
     sniff,
@@ -715,16 +747,18 @@ def insert_upsert_implementation(
 ):
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
-    if delimiter or quotechar or sniff or no_headers:
+    if (delimiter or quotechar or sniff or no_headers) and not tsv:
         csv = True
     if (nl + csv + tsv) >= 2:
         raise click.ClickException("Use just one of --nl, --csv or --tsv")
+    if (csv or tsv) and flatten:
+        raise click.ClickException("--flatten cannot be used with --csv or --tsv")
     if encoding and not (csv or tsv):
         raise click.ClickException("--encoding must be used with --csv or --tsv")
     if pk and len(pk) == 1:
         pk = pk[0]
     encoding = encoding or "utf-8-sig"
-    buffered = io.BufferedReader(json_file, buffer_size=4096)
+    buffered = io.BufferedReader(file, buffer_size=4096)
     decoded = io.TextIOWrapper(buffered, encoding=encoding)
     tracker = None
     if csv or tsv:
@@ -751,6 +785,10 @@ def insert_upsert_implementation(
             if detect_types:
                 tracker = TypeTracker()
                 docs = tracker.wrap(docs)
+    elif lines:
+        docs = ({"line": line.strip()} for line in decoded)
+    elif text:
+        docs = ({"text": decoded.read()},)
     else:
         try:
             if nl:
@@ -763,6 +801,31 @@ def insert_upsert_implementation(
             raise click.ClickException(
                 "Invalid JSON - use --csv for CSV or --tsv for TSV files"
             )
+        if flatten:
+            docs = (dict(_flatten(doc)) for doc in docs)
+
+    if convert:
+        variable = "row"
+        if lines:
+            variable = "line"
+        elif text:
+            variable = "text"
+        fn = _compile_code(convert, imports, variable=variable)
+        if lines:
+            docs = (fn(doc["line"]) for doc in docs)
+        elif text:
+            # Special case: this is allowed to be an iterable
+            text_value = list(docs)[0]["text"]
+            fn_return = fn(text_value)
+            if isinstance(fn_return, dict):
+                docs = [fn_return]
+            else:
+                try:
+                    docs = iter(fn_return)
+                except TypeError:
+                    raise click.ClickException("--convert must return dict or iterator")
+        else:
+            docs = (fn(doc) for doc in docs)
 
     extra_kwargs = {"ignore": ignore, "replace": replace, "truncate": truncate}
     if not_null:
@@ -771,22 +834,58 @@ def insert_upsert_implementation(
         extra_kwargs["defaults"] = dict(default)
     if upsert:
         extra_kwargs["upsert"] = upsert
+
     # Apply {"$base64": true, ...} decoding, if needed
     docs = (decode_base64_values(doc) for doc in docs)
+
     try:
         db[table].insert_all(
             docs, pk=pk, batch_size=batch_size, alter=alter, **extra_kwargs
         )
-    except sqlite3.OperationalError as e:
-        if e.args and "has no column named" in e.args[0]:
+    except Exception as e:
+        if (
+            isinstance(e, sqlite3.OperationalError)
+            and e.args
+            and "has no column named" in e.args[0]
+        ):
             raise click.ClickException(
                 "{}\n\nTry using --alter to add additional columns".format(e.args[0])
             )
-        raise
+        # If we can find sql= and parameters= arguments, show those
+        variables = _find_variables(e.__traceback__, ["sql", "parameters"])
+        if "sql" in variables and "parameters" in variables:
+            raise click.ClickException(
+                "{}\n\nsql = {}\nparameters = {}".format(
+                    str(e), variables["sql"], variables["parameters"]
+                )
+            )
+        else:
+            raise
     if tracker is not None:
         db[table].transform(types=tracker.types)
 
 
+def _flatten(d):
+    for key, value in d.items():
+        if isinstance(value, dict):
+            for key2, value2 in _flatten(value):
+                yield key + "_" + key2, value2
+        else:
+            yield key, value
+
+
+def _find_variables(tb, vars):
+    to_find = list(vars)
+    found = {}
+    for var in to_find:
+        if var in tb.tb_frame.f_locals:
+            vars.remove(var)
+            found[var] = tb.tb_frame.f_locals[var]
+    if vars and tb.tb_next:
+        found.update(_find_variables(tb.tb_next, vars))
+    return found
+
+
 @cli.command()
 @insert_upsert_options
 @click.option(
@@ -807,11 +906,16 @@ def insert_upsert_implementation(
 def insert(
     path,
     table,
-    json_file,
+    file,
     pk,
+    flatten,
     nl,
     csv,
     tsv,
+    lines,
+    text,
+    convert,
+    imports,
     delimiter,
     quotechar,
     sniff,
@@ -829,20 +933,40 @@ def insert(
     default,
 ):
     """
-    Insert records from JSON file into a table, creating the table if it
+    Insert records from FILE into a table, creating the table if it
     does not already exist.
 
-    Input should be a JSON array of objects, unless --nl or --csv is used.
+    By default the input is expected to be a JSON array of objects. Or:
+
+    \b
+    - Use --nl for newline-delimited JSON objects
+    - Use --csv or --tsv for comma-separated or tab-separated input
+    - Use --lines to write each incoming line to a column called "line"
+    - Use --text to write the entire input to a column called "text"
+
+    You can also use --convert to pass a fragment of Python code that will
+    be used to convert each input.
+
+    Your Python code will be passed a "row" variable representing the
+    imported row, and can return a modified row.
+
+    If you are using --lines your code will be passed a "line" variable,
+    and for --text an "text" variable.
     """
     try:
         insert_upsert_implementation(
             path,
             table,
-            json_file,
+            file,
             pk,
+            flatten,
             nl,
             csv,
             tsv,
+            lines,
+            text,
+            convert,
+            imports,
             delimiter,
             quotechar,
             sniff,
@@ -869,11 +993,16 @@ def insert(
 def upsert(
     path,
     table,
-    json_file,
+    file,
     pk,
+    flatten,
     nl,
     csv,
     tsv,
+    lines,
+    text,
+    convert,
+    imports,
     batch_size,
     delimiter,
     quotechar,
@@ -896,11 +1025,16 @@ def upsert(
         insert_upsert_implementation(
             path,
             table,
-            json_file,
+            file,
             pk,
+            flatten,
             nl,
             csv,
             tsv,
+            lines,
+            text,
+            convert,
+            imports,
             delimiter,
             quotechar,
             sniff,
@@ -911,6 +1045,7 @@ def upsert(
             not_null=not_null,
             default=default,
             encoding=encoding,
+            detect_types=detect_types,
             load_extension=load_extension,
             silent=silent,
         )
@@ -1143,6 +1278,11 @@ def query(
     multiple=True,
     help="Additional databases to attach - specify alias and filepath",
 )
+@click.option(
+    "--flatten",
+    is_flag=True,
+    help='Flatten nested JSON objects, so {"foo": {"bar": 1}} becomes {"foo_bar": 1}',
+)
 @output_options
 @click.option("-r", "--raw", is_flag=True, help="Raw output, first column of first row")
 @click.option(
@@ -1169,11 +1309,17 @@ def query(
     type=click.Path(file_okay=True, dir_okay=False, allow_dash=False),
     help="Save in-memory database to this file",
 )
+@click.option(
+    "--analyze",
+    is_flag=True,
+    help="Analyze resulting tables and output results",
+)
 @load_extension_option
 def memory(
     paths,
     sql,
     attach,
+    flatten,
     nl,
     arrays,
     csv,
@@ -1189,6 +1335,7 @@ def memory(
     schema,
     dump,
     save,
+    analyze,
     load_extension,
 ):
     """Execute SQL query against an in-memory database, optionally populated by imported data
@@ -1218,10 +1365,11 @@ def memory(
         sqlite-utils memory animals.csv --schema
     """
     db = sqlite_utils.Database(memory=True)
-    # If --dump or --save used but no paths detected, assume SQL query is a path:
-    if (dump or save or schema) and not paths:
+    # If --dump or --save or --analyze used but no paths detected, assume SQL query is a path:
+    if (dump or save or schema or analyze) and not paths:
         paths = [sql]
         sql = None
+    stem_counts = {}
     for i, path in enumerate(paths):
         # Path may have a :format suffix
         if ":" in path and path.rsplit(":", 1)[-1].upper() in Format.__members__:
@@ -1234,13 +1382,20 @@ def memory(
             csv_table = "stdin"
         else:
             csv_path = pathlib.Path(path)
-            csv_table = csv_path.stem
+            stem = csv_path.stem
+            if stem_counts.get(stem):
+                csv_table = "{}_{}".format(stem, stem_counts[stem])
+            else:
+                csv_table = stem
+            stem_counts[stem] = stem_counts.get(stem, 1) + 1
             csv_fp = csv_path.open("rb")
         rows, format_used = rows_from_file(csv_fp, format=format, encoding=encoding)
         tracker = None
         if format_used in (Format.CSV, Format.TSV) and not no_detect_types:
             tracker = TypeTracker()
             rows = tracker.wrap(rows)
+        if flatten:
+            rows = (dict(_flatten(row)) for row in rows)
         db[csv_table].insert_all(rows, alter=True)
         if tracker is not None:
             db[csv_table].transform(types=tracker.types)
@@ -1252,6 +1407,10 @@ def memory(
             if not db[view_name].exists():
                 db.create_view(view_name, "select * from [{}]".format(csv_table))
 
+    if analyze:
+        _analyze(db, tables=None, columns=None, save=False)
+        return
+
     if dump:
         for line in db.conn.iterdump():
             click.echo(line)
@@ -1328,6 +1487,7 @@ def _execute_query(
 @click.option(
     "--sql", "show_sql", is_flag=True, help="Show SQL query that would be run"
 )
+@click.option("--quote", is_flag=True, help="Apply FTS quoting rules to search term")
 @output_options
 @load_extension_option
 @click.pass_context
@@ -1338,6 +1498,7 @@ def search(
     q,
     order,
     show_sql,
+    quote,
     column,
     limit,
     nl,
@@ -1373,21 +1534,31 @@ def search(
     if show_sql:
         click.echo(sql)
         return
-    ctx.invoke(
-        query,
-        path=path,
-        sql=sql,
-        nl=nl,
-        arrays=arrays,
-        csv=csv,
-        tsv=tsv,
-        no_headers=no_headers,
-        table=table,
-        fmt=fmt,
-        json_cols=json_cols,
-        param=[("query", q)],
-        load_extension=load_extension,
-    )
+    if quote:
+        q = db.quote_fts(q)
+    try:
+        ctx.invoke(
+            query,
+            path=path,
+            sql=sql,
+            nl=nl,
+            arrays=arrays,
+            csv=csv,
+            tsv=tsv,
+            no_headers=no_headers,
+            table=table,
+            fmt=fmt,
+            json_cols=json_cols,
+            param=[("query", q)],
+            load_extension=load_extension,
+        )
+    except click.ClickException as e:
+        if "malformed MATCH expression" in str(e) or "unterminated string" in str(e):
+            raise click.ClickException(
+                "{}\n\nTry running this again with the --quote option".format(str(e))
+            )
+        else:
+            raise
 
 
 @cli.command()
@@ -1742,9 +1913,27 @@ def extract(
 @click.option("--replace", is_flag=True, help="Replace files with matching primary key")
 @click.option("--upsert", is_flag=True, help="Upsert files with matching primary key")
 @click.option("--name", type=str, help="File name to use")
+@click.option("--text", is_flag=True, help="Store file content as TEXT, not BLOB")
+@click.option(
+    "--encoding",
+    help="Character encoding for input, defaults to utf-8",
+)
+@click.option("-s", "--silent", is_flag=True, help="Don't show a progress bar")
 @load_extension_option
 def insert_files(
-    path, table, file_or_dir, column, pk, alter, replace, upsert, name, load_extension
+    path,
+    table,
+    file_or_dir,
+    column,
+    pk,
+    alter,
+    replace,
+    upsert,
+    name,
+    text,
+    encoding,
+    silent,
+    load_extension,
 ):
     """
     Insert one or more files using BLOB columns in the specified table
@@ -1762,7 +1951,10 @@ def insert_files(
         --pk name
     """
     if not column:
-        column = ["path:path", "content:content", "size:size"]
+        if text:
+            column = ["path:path", "content_text:content_text", "size:size"]
+        else:
+            column = ["path:path", "content:content", "size:size"]
         if not pk:
             pk = "path"
 
@@ -1781,12 +1973,21 @@ def insert_files(
     # Load all paths so we can show a progress bar
     paths_and_relative_paths = list(yield_paths_and_relative_paths())
 
-    with click.progressbar(paths_and_relative_paths) as bar:
+    with progressbar(paths_and_relative_paths, silent=silent) as bar:
 
         def to_insert():
             for path, relative_path in bar:
                 row = {}
-                lookups = FILE_COLUMNS
+                # content_text is special case as it considers 'encoding'
+
+                def _content_text(p):
+                    resolved = p.resolve()
+                    try:
+                        return resolved.read_text(encoding=encoding)
+                    except UnicodeDecodeError as e:
+                        raise UnicodeDecodeErrorForPath(e, resolved)
+
+                lookups = dict(FILE_COLUMNS, content_text=_content_text)
                 if path == "-":
                     stdin_data = sys.stdin.buffer.read()
                     # We only support a subset of columns for this case
@@ -1794,6 +1995,9 @@ def insert_files(
                         "name": lambda p: name or "-",
                         "path": lambda p: name or "-",
                         "content": lambda p: stdin_data,
+                        "content_text": lambda p: stdin_data.decode(
+                            encoding or "utf-8"
+                        ),
                         "sha256": lambda p: hashlib.sha256(stdin_data).hexdigest(),
                         "md5": lambda p: hashlib.md5(stdin_data).hexdigest(),
                         "size": lambda p: len(stdin_data),
@@ -1819,9 +2023,16 @@ def insert_files(
 
         db = sqlite_utils.Database(path)
         _load_extensions(db, load_extension)
-        with db.conn:
-            db[table].insert_all(
-                to_insert(), pk=pk, alter=alter, replace=replace, upsert=upsert
+        try:
+            with db.conn:
+                db[table].insert_all(
+                    to_insert(), pk=pk, alter=alter, replace=replace, upsert=upsert
+                )
+        except UnicodeDecodeErrorForPath as e:
+            raise click.ClickException(
+                UNICODE_ERROR.format(
+                    "Could not read file '{}' as text\n\n{}".format(e.path, e.exception)
+                )
             )
 
 
@@ -1852,6 +2063,10 @@ def analyze_tables(
     "Analyze the columns in one or more tables"
     db = sqlite_utils.Database(path)
     _load_extensions(db, load_extension)
+    _analyze(db, tables, columns, save)
+
+
+def _analyze(db, tables, columns, save):
     if not tables:
         tables = db.table_names()
     todo = []
@@ -1903,6 +2118,159 @@ def analyze_tables(
         click.echo(details)
 
 
+def _generate_convert_help():
+    help = textwrap.dedent(
+        """
+    Convert columns using Python code you supply. For example:
+
+    \b
+    $ sqlite-utils convert my.db mytable mycolumn \\
+        '"\\n".join(textwrap.wrap(value, 10))' \\
+        --import=textwrap
+
+    "value" is a variable with the column value to be converted.
+
+    Use "-" for CODE to read Python code from standard input.
+
+    The following common operations are available as recipe functions:
+    """
+    ).strip()
+    recipe_names = [
+        n for n in dir(recipes) if not n.startswith("_") and n not in ("json", "parser")
+    ]
+    for name in recipe_names:
+        fn = getattr(recipes, name)
+        help += "\n\nr.{}{}\n\n  {}".format(
+            name, str(inspect.signature(fn)), fn.__doc__
+        )
+    help += "\n\n"
+    help += textwrap.dedent(
+        """
+    You can use these recipes like so:
+
+    \b
+    $ sqlite-utils convert my.db mytable mycolumn \\
+        'r.jsonsplit(value, delimiter=":")'
+    """
+    ).strip()
+    return help
+
+
+@cli.command(help=_generate_convert_help())
+@click.argument(
+    "db_path",
+    type=click.Path(file_okay=True, dir_okay=False, allow_dash=False),
+    required=True,
+)
+@click.argument("table", type=str)
+@click.argument("columns", type=str, nargs=-1, required=True)
+@click.argument("code", type=str)
+@click.option(
+    "--import", "imports", type=str, multiple=True, help="Python modules to import"
+)
+@click.option(
+    "--dry-run", is_flag=True, help="Show results of running this against first 10 rows"
+)
+@click.option(
+    "--multi", is_flag=True, help="Populate columns for keys in returned dictionary"
+)
+@click.option("--where", help="Optional where clause")
+@click.option(
+    "-p",
+    "--param",
+    multiple=True,
+    type=(str, str),
+    help="Named :parameters for where clause",
+)
+@click.option("--output", help="Optional separate column to populate with the output")
+@click.option(
+    "--output-type",
+    help="Column type to use for the output column",
+    default="text",
+    type=click.Choice(["integer", "float", "blob", "text"]),
+)
+@click.option("--drop", is_flag=True, help="Drop original column afterwards")
+@click.option("-s", "--silent", is_flag=True, help="Don't show a progress bar")
+def convert(
+    db_path,
+    table,
+    columns,
+    code,
+    imports,
+    dry_run,
+    multi,
+    where,
+    param,
+    output,
+    output_type,
+    drop,
+    silent,
+):
+    sqlite3.enable_callback_tracebacks(True)
+    db = sqlite_utils.Database(db_path)
+    if output is not None and len(columns) > 1:
+        raise click.ClickException("Cannot use --output with more than one column")
+    if multi and len(columns) > 1:
+        raise click.ClickException("Cannot use --multi with more than one column")
+    if drop and not (output or multi):
+        raise click.ClickException("--drop can only be used with --output or --multi")
+    if code == "-":
+        # Read code from standard input
+        code = sys.stdin.read()
+    where_args = dict(param) if param else []
+    # Compile the code into a function body called fn(value)
+    try:
+        fn = _compile_code(code, imports)
+    except SyntaxError as e:
+        raise click.ClickException(
+            "Syntax error in code:\n\n{}\n\n{}".format(
+                textwrap.indent(e.text.strip(), "    "), e.msg
+            )
+        )
+    if dry_run:
+        # Pull first 20 values for first column and preview them
+        db.conn.create_function("preview_transform", 1, lambda v: fn(v) if v else v)
+        sql = """
+            select
+                [{column}] as value,
+                preview_transform([{column}]) as preview
+            from [{table}]{where} limit 10
+        """.format(
+            column=columns[0],
+            table=table,
+            where=" where {}".format(where) if where is not None else "",
+        )
+        for row in db.conn.execute(sql, where_args).fetchall():
+            click.echo(str(row[0]))
+            click.echo(" --- becomes:")
+            click.echo(str(row[1]))
+            click.echo()
+        count = db[table].count_where(
+            where=where,
+            where_args=where_args,
+        )
+        click.echo("Would affect {} row{}".format(count, "" if count == 1 else "s"))
+    else:
+        try:
+            db[table].convert(
+                columns,
+                fn,
+                where=where,
+                where_args=where_args,
+                output=output,
+                output_type=output_type,
+                drop=drop,
+                multi=multi,
+                show_progress=not silent,
+            )
+        except BadMultiValues as e:
+            raise click.ClickException(
+                "When using --multi code must return a Python dictionary - returned: {}".format(
+                    repr(e.values)
+                )
+            )
+
+
 def _render_common(title, values):
     if values is None:
         return ""
@@ -1912,6 +2280,12 @@ def _render_common(title, values):
     return "\n".join(lines)
 
 
+class UnicodeDecodeErrorForPath(Exception):
+    def __init__(self, exception, path):
+        self.exception = exception
+        self.path = path
+
+
 FILE_COLUMNS = {
     "name": lambda p: p.name,
     "path": lambda p: str(p),
@@ -1953,6 +2327,9 @@ def output_rows(iterator, headers, nl, arrays, json_cols):
         )
         yield line
         first = False
+    if first:
+        # We didn't output any rows, so yield the empty list
+        yield "[]"
 
 
 def maybe_json(value):
