@@ -11,20 +11,6 @@ import textwrap
 from .utils import collapse_whitespace
 
 
-CREATE_TABLES = """
-create table Gosh (c1 text, c2 text, c3 text);
-create table Gosh2 (c1 text, c2 text, c3 text);
-"""
-
-
-@pytest.fixture
-def db_path(tmpdir):
-    path = str(tmpdir / "test.db")
-    db = sqlite3.connect(path)
-    db.executescript(CREATE_TABLES)
-    return path
-
-
 @pytest.mark.parametrize(
     "options",
     (
@@ -556,8 +542,7 @@ def test_optimize(db_path, tables):
     assert 0 == result.exit_code
 
 
-@pytest.mark.parametrize("tables", ([], ["fts4_table"], ["fts5_table"]))
-def test_rebuild_fts(db_path, tables):
+def test_rebuild_fts_fixes_docsize_error(db_path):
     db = Database(db_path, recursive_triggers=False)
     records = [
         {
@@ -568,335 +553,22 @@ def test_rebuild_fts(db_path, tables):
         for i in range(10000)
     ]
     with db.conn:
-        db["fts4_table"].insert_all(records, pk="c1")
         db["fts5_table"].insert_all(records, pk="c1")
-        db["fts4_table"].enable_fts(
-            ["c1", "c2", "c3"], fts_version="FTS4", create_triggers=True
-        )
         db["fts5_table"].enable_fts(
             ["c1", "c2", "c3"], fts_version="FTS5", create_triggers=True
         )
     # Search should work
-    assert list(db["fts4_table"].search("verb1"))
     assert list(db["fts5_table"].search("verb1"))
-    # Deleting _fts_segments to break FTS4
-    with db.conn:
-        db["fts4_table_fts_segments"].delete_where()
-    # Now this should error:
-    with pytest.raises(sqlite3.DatabaseError):
-        list(db["fts4_table"].search("verb1"))
     # Replicate docsize error from this issue for FTS5
     # https://github.com/simonw/sqlite-utils/issues/149
     assert db["fts5_table_fts_docsize"].count == 10000
     db["fts5_table"].insert_all(records, replace=True)
     assert db["fts5_table"].count == 10000
     assert db["fts5_table_fts_docsize"].count == 20000
-    # Running rebuild-fts should fix both issues
-    print(["rebuild-fts", db_path] + tables)
-    result = CliRunner().invoke(cli.cli, ["rebuild-fts", db_path] + tables)
+    # Running rebuild-fts should fix this
+    result = CliRunner().invoke(cli.cli, ["rebuild-fts", db_path, "fts5_table"])
     assert 0 == result.exit_code
-    fixed_tables = tables or ["fts4_table", "fts5_table"]
-    if "fts4_table" in fixed_tables:
-        assert list(db["fts4_table"].search("verb1"))
-    else:
-        with pytest.raises(sqlite3.DatabaseError):
-            list(db["fts4_table"].search("verb1"))
-    if "fts5_table" in fixed_tables:
-        assert db["fts5_table_fts_docsize"].count == 10000
-    else:
-        assert db["fts5_table_fts_docsize"].count == 20000
-
-
-def test_insert_simple(tmpdir):
-    json_path = str(tmpdir / "dog.json")
-    db_path = str(tmpdir / "dogs.db")
-    open(json_path, "w").write(json.dumps({"name": "Cleo", "age": 4}))
-    result = CliRunner().invoke(cli.cli, ["insert", db_path, "dogs", json_path])
-    assert 0 == result.exit_code
-    assert [{"age": 4, "name": "Cleo"}] == list(
-        Database(db_path).query("select * from dogs")
-    )
-    db = Database(db_path)
-    assert ["dogs"] == db.table_names()
-    assert [] == db["dogs"].indexes
-
-
-def test_insert_from_stdin(tmpdir):
-    db_path = str(tmpdir / "dogs.db")
-    result = CliRunner().invoke(
-        cli.cli,
-        ["insert", db_path, "dogs", "-"],
-        input=json.dumps({"name": "Cleo", "age": 4}),
-    )
-    assert 0 == result.exit_code
-    assert [{"age": 4, "name": "Cleo"}] == list(
-        Database(db_path).query("select * from dogs")
-    )
-
-
-def test_insert_invalid_json_error(tmpdir):
-    db_path = str(tmpdir / "dogs.db")
-    result = CliRunner().invoke(
-        cli.cli,
-        ["insert", db_path, "dogs", "-"],
-        input="name,age\nCleo,4",
-    )
-    assert result.exit_code == 1
-    assert (
-        result.output
-        == "Error: Invalid JSON - use --csv for CSV or --tsv for TSV files\n"
-    )
-
-
-def test_insert_with_primary_key(db_path, tmpdir):
-    json_path = str(tmpdir / "dog.json")
-    open(json_path, "w").write(json.dumps({"id": 1, "name": "Cleo", "age": 4}))
-    result = CliRunner().invoke(
-        cli.cli, ["insert", db_path, "dogs", json_path, "--pk", "id"]
-    )
-    assert 0 == result.exit_code
-    assert [{"id": 1, "age": 4, "name": "Cleo"}] == list(
-        Database(db_path).query("select * from dogs")
-    )
-    db = Database(db_path)
-    assert ["id"] == db["dogs"].pks
-
-
-def test_insert_multiple_with_primary_key(db_path, tmpdir):
-    json_path = str(tmpdir / "dogs.json")
-    dogs = [{"id": i, "name": "Cleo {}".format(i), "age": i + 3} for i in range(1, 21)]
-    open(json_path, "w").write(json.dumps(dogs))
-    result = CliRunner().invoke(
-        cli.cli, ["insert", db_path, "dogs", json_path, "--pk", "id"]
-    )
-    assert 0 == result.exit_code
-    db = Database(db_path)
-    assert dogs == list(db.query("select * from dogs order by id"))
-    assert ["id"] == db["dogs"].pks
-
-
-def test_insert_multiple_with_compound_primary_key(db_path, tmpdir):
-    json_path = str(tmpdir / "dogs.json")
-    dogs = [
-        {"breed": "mixed", "id": i, "name": "Cleo {}".format(i), "age": i + 3}
-        for i in range(1, 21)
-    ]
-    open(json_path, "w").write(json.dumps(dogs))
-    result = CliRunner().invoke(
-        cli.cli, ["insert", db_path, "dogs", json_path, "--pk", "id", "--pk", "breed"]
-    )
-    assert 0 == result.exit_code
-    db = Database(db_path)
-    assert dogs == list(db.query("select * from dogs order by breed, id"))
-    assert {"breed", "id"} == set(db["dogs"].pks)
-    assert (
-        "CREATE TABLE [dogs] (\n"
-        "   [breed] TEXT,\n"
-        "   [id] INTEGER,\n"
-        "   [name] TEXT,\n"
-        "   [age] INTEGER,\n"
-        "   PRIMARY KEY ([id], [breed])\n"
-        ")"
-    ) == db["dogs"].schema
-
-
-def test_insert_not_null_default(db_path, tmpdir):
-    json_path = str(tmpdir / "dogs.json")
-    dogs = [
-        {"id": i, "name": "Cleo {}".format(i), "age": i + 3, "score": 10}
-        for i in range(1, 21)
-    ]
-    open(json_path, "w").write(json.dumps(dogs))
-    result = CliRunner().invoke(
-        cli.cli,
-        ["insert", db_path, "dogs", json_path, "--pk", "id"]
-        + ["--not-null", "name", "--not-null", "age"]
-        + ["--default", "score", "5", "--default", "age", "1"],
-    )
-    assert 0 == result.exit_code
-    db = Database(db_path)
-    assert (
-        "CREATE TABLE [dogs] (\n"
-        "   [id] INTEGER PRIMARY KEY,\n"
-        "   [name] TEXT NOT NULL,\n"
-        "   [age] INTEGER NOT NULL DEFAULT '1',\n"
-        "   [score] INTEGER DEFAULT '5'\n)"
-    ) == db["dogs"].schema
-
-
-def test_insert_binary_base64(db_path):
-    result = CliRunner().invoke(
-        cli.cli,
-        ["insert", db_path, "files", "-"],
-        input=r'{"content": {"$base64": true, "encoded": "aGVsbG8="}}',
-    )
-    assert 0 == result.exit_code, result.output
-    db = Database(db_path)
-    actual = list(db.query("select content from files"))
-    assert actual == [{"content": b"hello"}]
-
-
-def test_insert_newline_delimited(db_path):
-    result = CliRunner().invoke(
-        cli.cli,
-        ["insert", db_path, "from_json_nl", "-", "--nl"],
-        input='{"foo": "bar", "n": 1}\n{"foo": "baz", "n": 2}',
-    )
-    assert 0 == result.exit_code, result.output
-    db = Database(db_path)
-    assert [
-        {"foo": "bar", "n": 1},
-        {"foo": "baz", "n": 2},
-    ] == list(db.query("select foo, n from from_json_nl"))
-
-
-def test_insert_ignore(db_path, tmpdir):
-    db = Database(db_path)
-    db["dogs"].insert({"id": 1, "name": "Cleo"}, pk="id")
-    json_path = str(tmpdir / "dogs.json")
-    open(json_path, "w").write(json.dumps([{"id": 1, "name": "Bailey"}]))
-    # Should raise error without --ignore
-    result = CliRunner().invoke(
-        cli.cli, ["insert", db_path, "dogs", json_path, "--pk", "id"]
-    )
-    assert 0 != result.exit_code, result.output
-    # If we use --ignore it should run OK
-    result = CliRunner().invoke(
-        cli.cli, ["insert", db_path, "dogs", json_path, "--pk", "id", "--ignore"]
-    )
-    assert 0 == result.exit_code, result.output
-    # ... but it should actually have no effect
-    assert [{"id": 1, "name": "Cleo"}] == list(db.query("select * from dogs"))
-
-
-@pytest.mark.parametrize(
-    "content,options",
-    [
-        ("foo\tbar\tbaz\n1\t2\tcat,dog", ["--tsv"]),
-        ('foo,bar,baz\n1,2,"cat,dog"', ["--csv"]),
-        ('foo;bar;baz\n1;2;"cat,dog"', ["--csv", "--delimiter", ";"]),
-        # --delimiter implies --csv:
-        ('foo;bar;baz\n1;2;"cat,dog"', ["--delimiter", ";"]),
-        ("foo,bar,baz\n1,2,|cat,dog|", ["--csv", "--quotechar", "|"]),
-        ("foo,bar,baz\n1,2,|cat,dog|", ["--quotechar", "|"]),
-    ],
-)
-def test_insert_csv_tsv(content, options, db_path, tmpdir):
-    db = Database(db_path)
-    file_path = str(tmpdir / "insert.csv-tsv")
-    open(file_path, "w").write(content)
-    result = CliRunner().invoke(
-        cli.cli,
-        ["insert", db_path, "data", file_path] + options,
-        catch_exceptions=False,
-    )
-    assert 0 == result.exit_code
-    assert [{"foo": "1", "bar": "2", "baz": "cat,dog"}] == list(db["data"].rows)
-
-
-@pytest.mark.parametrize(
-    "options",
-    (
-        ["--tsv", "--nl"],
-        ["--tsv", "--csv"],
-        ["--csv", "--nl"],
-        ["--csv", "--nl", "--tsv"],
-    ),
-)
-def test_only_allow_one_of_nl_tsv_csv(options, db_path, tmpdir):
-    file_path = str(tmpdir / "insert.csv-tsv")
-    open(file_path, "w").write("foo")
-    result = CliRunner().invoke(
-        cli.cli, ["insert", db_path, "data", file_path] + options
-    )
-    assert 0 != result.exit_code
-    assert "Error: Use just one of --nl, --csv or --tsv" == result.output.strip()
-
-
-def test_insert_replace(db_path, tmpdir):
-    test_insert_multiple_with_primary_key(db_path, tmpdir)
-    json_path = str(tmpdir / "insert-replace.json")
-    db = Database(db_path)
-    assert 20 == db["dogs"].count
-    insert_replace_dogs = [
-        {"id": 1, "name": "Insert replaced 1", "age": 4},
-        {"id": 2, "name": "Insert replaced 2", "age": 4},
-        {"id": 21, "name": "Fresh insert 21", "age": 6},
-    ]
-    open(json_path, "w").write(json.dumps(insert_replace_dogs))
-    result = CliRunner().invoke(
-        cli.cli, ["insert", db_path, "dogs", json_path, "--pk", "id", "--replace"]
-    )
-    assert 0 == result.exit_code, result.output
-    assert 21 == db["dogs"].count
-    assert (
-        list(db.query("select * from dogs where id in (1, 2, 21) order by id"))
-        == insert_replace_dogs
-    )
-
-
-def test_insert_truncate(db_path):
-    result = CliRunner().invoke(
-        cli.cli,
-        ["insert", db_path, "from_json_nl", "-", "--nl", "--batch-size=1"],
-        input='{"foo": "bar", "n": 1}\n{"foo": "baz", "n": 2}',
-    )
-    assert 0 == result.exit_code, result.output
-    db = Database(db_path)
-    assert [
-        {"foo": "bar", "n": 1},
-        {"foo": "baz", "n": 2},
-    ] == list(db.query("select foo, n from from_json_nl"))
-    # Truncate and insert new rows
-    result = CliRunner().invoke(
-        cli.cli,
-        [
-            "insert",
-            db_path,
-            "from_json_nl",
-            "-",
-            "--nl",
-            "--truncate",
-            "--batch-size=1",
-        ],
-        input='{"foo": "bam", "n": 3}\n{"foo": "bat", "n": 4}',
-    )
-    assert 0 == result.exit_code, result.output
-    assert [
-        {"foo": "bam", "n": 3},
-        {"foo": "bat", "n": 4},
-    ] == list(db.query("select foo, n from from_json_nl"))
-
-
-def test_insert_alter(db_path, tmpdir):
-    result = CliRunner().invoke(
-        cli.cli,
-        ["insert", db_path, "from_json_nl", "-", "--nl"],
-        input='{"foo": "bar", "n": 1}\n{"foo": "baz", "n": 2}',
-    )
-    assert 0 == result.exit_code, result.output
-    # Should get an error with incorrect shaped additional data
-    result = CliRunner().invoke(
-        cli.cli,
-        ["insert", db_path, "from_json_nl", "-", "--nl"],
-        input='{"foo": "bar", "baz": 5}',
-    )
-    assert 0 != result.exit_code, result.output
-    # If we run it again with --alter it should work correctly
-    result = CliRunner().invoke(
-        cli.cli,
-        ["insert", db_path, "from_json_nl", "-", "--nl", "--alter"],
-        input='{"foo": "bar", "baz": 5}',
-    )
-    assert 0 == result.exit_code, result.output
-    # Soundness check the database itself
-    db = Database(db_path)
-    assert {"foo": str, "n": int, "baz": int} == db["from_json_nl"].columns_dict
-    assert [
-        {"foo": "bar", "n": 1, "baz": None},
-        {"foo": "baz", "n": 2, "baz": None},
-        {"foo": "bar", "baz": 5, "n": None},
-    ] == list(db.query("select foo, n, baz from from_json_nl"))
+    assert db["fts5_table_fts_docsize"].count == 10000
 
 
 @pytest.mark.parametrize(
@@ -966,6 +638,13 @@ def test_query_json(db_path, sql, args, expected):
     assert expected == result.output.strip()
 
 
+def test_query_json_empty(db_path):
+    result = CliRunner().invoke(
+        cli.cli, [db_path, "select * from sqlite_master where 0"]
+    )
+    assert result.output.strip() == "[]"
+
+
 LOREM_IPSUM_COMPRESSED = (
     b"x\x9c\xed\xd1\xcdq\x03!\x0c\x05\xe0\xbb\xabP\x01\x1eW\x91\xdc|M\x01\n\xc8\x8e"
     b"f\xf83H\x1e\x97\x1f\x91M\x8e\xe9\xe0\xdd\x96\x05\x84\xf4\xbek\x9fRI\xc7\xf2J"
@@ -1197,6 +876,21 @@ def test_upsert(db_path, tmpdir):
     ]
 
 
+def test_upsert_flatten(tmpdir):
+    db_path = str(tmpdir / "flat.db")
+    db = Database(db_path)
+    db["upsert_me"].insert({"id": 1, "name": "Example"}, pk="id")
+    result = CliRunner().invoke(
+        cli.cli,
+        ["upsert", db_path, "upsert_me", "-", "--flatten", "--pk", "id", "--alter"],
+        input=json.dumps({"id": 1, "nested": {"two": 2}}),
+    )
+    assert result.exit_code == 0
+    assert list(db.query("select * from upsert_me")) == [
+        {"id": 1, "name": "Example", "nested_two": 2}
+    ]
+
+
 def test_upsert_alter(db_path, tmpdir):
     json_path = str(tmpdir / "dogs.json")
     db = Database(db_path)
@@ -1213,7 +907,11 @@ def test_upsert_alter(db_path, tmpdir):
         cli.cli, ["upsert", db_path, "dogs", json_path, "--pk", "id"]
     )
     assert 1 == result.exit_code
-    assert "no such column: age" == str(result.exception)
+    assert (
+        "Error: no such column: age\n\n"
+        "sql = UPDATE [dogs] SET [age] = ? WHERE [id] = ?\n"
+        "parameters = [5, 1]"
+    ) == result.output.strip()
     # Should succeed with --alter
     result = CliRunner().invoke(
         cli.cli, ["upsert", db_path, "dogs", json_path, "--pk", "id", "--alter"]
@@ -1947,6 +1645,25 @@ def test_search(tmpdir, fts, extra_arg, expected):
     assert result.output.replace("\r", "") == expected
 
 
+def test_search_quote(tmpdir):
+    db_path = str(tmpdir / "test.db")
+    db = Database(db_path)
+    db["creatures"].insert({"name": "dog."}).enable_fts(["name"])
+    # Without --quote should return an error
+    error_result = CliRunner().invoke(cli.cli, ["search", db_path, "creatures", 'dog"'])
+    assert error_result.exit_code == 1
+    assert error_result.output == (
+        "Error: unterminated string\n\n"
+        "Try running this again with the --quote option\n"
+    )
+    # With --quote it should work
+    result = CliRunner().invoke(
+        cli.cli, ["search", db_path, "creatures", 'dog"', "--quote"]
+    )
+    assert result.exit_code == 0
+    assert result.output.strip() == '[{"rowid": 1, "name": "dog."}]'
+
+
 def test_indexes(tmpdir):
     db_path = str(tmpdir / "test.db")
     db = Database(db_path)
@@ -2032,7 +1749,11 @@ _TRIGGERS_EXPECTED = (
 
 @pytest.mark.parametrize(
     "extra_args,expected",
-    [([], _TRIGGERS_EXPECTED), (["articles"], _TRIGGERS_EXPECTED), (["counter"], "")],
+    [
+        ([], _TRIGGERS_EXPECTED),
+        (["articles"], _TRIGGERS_EXPECTED),
+        (["counter"], "[]\n"),
+    ],
 )
 def test_triggers(tmpdir, extra_args, expected):
     db_path = str(tmpdir / "test.db")
@@ -2139,25 +1860,27 @@ def test_long_csv_column_value(tmpdir):
 
 
 @pytest.mark.parametrize(
-    "args",
+    "args,tsv",
     (
-        ["--csv", "--no-headers"],
-        ["--no-headers"],
+        (["--csv", "--no-headers"], False),
+        (["--no-headers"], False),
+        (["--tsv", "--no-headers"], True),
     ),
 )
-def test_csv_import_no_headers(tmpdir, args):
+def test_import_no_headers(tmpdir, args, tsv):
     db_path = str(tmpdir / "test.db")
     csv_path = str(tmpdir / "test.csv")
     csv_file = open(csv_path, "w")
-    csv_file.write("Cleo,Dog,5\n")
-    csv_file.write("Tracy,Spider,7\n")
+    sep = "\t" if tsv else ","
+    csv_file.write("Cleo{sep}Dog{sep}5\n".format(sep=sep))
+    csv_file.write("Tracy{sep}Spider{sep}7\n".format(sep=sep))
     csv_file.close()
     result = CliRunner().invoke(
         cli.cli,
         ["insert", db_path, "creatures", csv_path] + args,
         catch_exceptions=False,
     )
-    assert result.exit_code == 0
+    assert result.exit_code == 0, result.output
     db = Database(db_path)
     schema = db["creatures"].schema
     assert schema == (
@@ -2249,3 +1972,48 @@ def test_insert_detect_types(tmpdir, option_or_env_var):
             _test()
     else:
         _test()
+
+
+@pytest.mark.parametrize("option", ("-d", "--detect-types"))
+def test_upsert_detect_types(tmpdir, option):
+    db_path = str(tmpdir / "test.db")
+    data = "id,name,age,weight\n1,Cleo,6,45.5\n2,Dori,1,3.5"
+    result = CliRunner().invoke(
+        cli.cli,
+        ["upsert", db_path, "creatures", "-", "--csv", "--pk", "id"] + [option],
+        catch_exceptions=False,
+        input=data,
+    )
+    assert result.exit_code == 0
+    db = Database(db_path)
+    assert list(db["creatures"].rows) == [
+        {"id": 1, "name": "Cleo", "age": 6, "weight": 45.5},
+        {"id": 2, "name": "Dori", "age": 1, "weight": 3.5},
+    ]
+
+
+@pytest.mark.parametrize(
+    "input,expected",
+    (
+        ({"foo": {"bar": 1}}, {"foo_bar": 1}),
+        ({"foo": {"bar": [1, 2, {"baz": 3}]}}, {"foo_bar": [1, 2, {"baz": 3}]}),
+        ({"foo": {"bar": 1, "baz": {"three": 3}}}, {"foo_bar": 1, "foo_baz_three": 3}),
+    ),
+)
+def test_flatten_helper(input, expected):
+    assert dict(cli._flatten(input)) == expected
+
+
+def test_integer_overflow_error(tmpdir):
+    db_path = str(tmpdir / "test.db")
+    result = CliRunner().invoke(
+        cli.cli,
+        ["insert", db_path, "items", "-"],
+        input=json.dumps({"bignumber": 34223049823094832094802398430298048240}),
+    )
+    assert result.exit_code == 1
+    assert result.output == (
+        "Error: Python int too large to convert to SQLite INTEGER\n\n"
+        "sql = INSERT INTO [items] ([bignumber]) VALUES (?);\n"
+        "parameters = [34223049823094832094802398430298048240]\n"
+    )
