@@ -4,6 +4,7 @@ from sqlite_utils.db import (
     DescIndex,
     AlterError,
     NoObviousTable,
+    OperationalError,
     ForeignKey,
     Table,
     View,
@@ -20,9 +21,9 @@ import uuid
 from .utils import collapse_whitespace
 
 try:
-    import pandas as pd
+    import pandas as pd  # type: ignore
 except ImportError:
-    pd = None
+    pd = None  # type: ignore
 
 
 def test_create_table(fresh_db):
@@ -143,6 +144,7 @@ def test_create_table_with_not_null(fresh_db):
             [{"name": "memoryview", "type": "BLOB"}],
         ),
         ({"uuid": uuid.uuid4()}, [{"name": "uuid", "type": "TEXT"}]),
+        ({"foo[bar]": 1}, [{"name": "foo_bar_", "type": "INTEGER"}]),
     ),
 )
 def test_create_table_from_example(fresh_db, example, expected_columns):
@@ -474,6 +476,18 @@ def test_index_foreign_keys(fresh_db):
     assert [["breed_id"]] == [i.columns for i in fresh_db["dogs"].indexes]
 
 
+def test_index_foreign_keys_if_index_name_is_already_used(fresh_db):
+    # https://github.com/simonw/sqlite-utils/issues/335
+    test_add_foreign_key_guess_table(fresh_db)
+    # Add index with a name that will conflict with index_foreign_keys()
+    fresh_db["dogs"].create_index(["name"], index_name="idx_dogs_breed_id")
+    fresh_db.index_foreign_keys()
+    assert {(idx.name, tuple(idx.columns)) for idx in fresh_db["dogs"].indexes} == {
+        ("idx_dogs_breed_id_2", ("breed_id",)),
+        ("idx_dogs_breed_id", ("name",)),
+    }
+
+
 @pytest.mark.parametrize(
     "extra_data,expected_new_columns",
     [
@@ -512,13 +526,6 @@ def test_insert_row_alter_table(
     ]
 
 
-def test_insert_row_alter_table_invalid_column_characters(fresh_db):
-    table = fresh_db["table"]
-    table.insert({"foo": "bar"}).last_pk
-    with pytest.raises(AssertionError):
-        table.insert({"foo": "baz", "new_col[abc]": 1.2}, alter=True)
-
-
 def test_add_missing_columns_case_insensitive(fresh_db):
     table = fresh_db["foo"]
     table.insert({"id": 1, "name": "Cleo"}, pk="id")
@@ -752,6 +759,21 @@ def test_create_index_desc(fresh_db):
     )
 
 
+def test_create_index_find_unique_name(fresh_db):
+    table = fresh_db["t"]
+    table.insert({"id": 1})
+    table.create_index(["id"])
+    # Without find_unique_name should error
+    with pytest.raises(OperationalError, match="index idx_t_id already exists"):
+        table.create_index(["id"])
+    # With find_unique_name=True it should work
+    table.create_index(["id"], find_unique_name=True)
+    table.create_index(["id"], find_unique_name=True)
+    # Should have three now
+    index_names = {idx.name for idx in table.indexes}
+    assert index_names == {"idx_t_id", "idx_t_id_2", "idx_t_id_3"}
+
+
 @pytest.mark.parametrize(
     "data_structure",
     (
@@ -1012,3 +1034,30 @@ def test_create_with_nested_bytes(fresh_db):
 )
 def test_quote(fresh_db, input, expected):
     assert fresh_db.quote(input) == expected
+
+
+@pytest.mark.parametrize(
+    "columns,expected_sql_middle",
+    (
+        (
+            {"id": int},
+            "[id] INTEGER",
+        ),
+        (
+            {"col": dict},
+            "[col] TEXT",
+        ),
+        (
+            {"col": tuple},
+            "[col] TEXT",
+        ),
+        (
+            {"col": list},
+            "[col] TEXT",
+        ),
+    ),
+)
+def test_create_table_sql(fresh_db, columns, expected_sql_middle):
+    sql = fresh_db.create_table_sql("t", columns)
+    middle = sql.split("(")[1].split(")")[0].strip()
+    assert middle == expected_sql_middle
